#include <math.h>
#include <memory.h>
#include <string.h>

#include "SimpleMOCoutput.logic.h"

#ifdef _WIN32
// strptime is missing from the MS world... use a local copy
#include "strptime.h"
#define SNPRINTF _snprintf
#else
#include <time.h>
#define SNPRINTF snprintf
#endif

#define CONTROLLED_RUN(x)                                                                           \
        {                                                                                           \
            int result = x;                                                                         \
            if (0 != result) {                                                                      \
                SNPRINTF(g_exceptionBuffer, sizeof(g_exceptionBuffer), "Failed to execute:\n%s\n(result:%d)\n", #x, result); \
                g_pUtils->pfnThrowException(g_exceptionBuffer);                                     \
                return;                                                                             \
            }                                                                                       \
        }


extern Utils *g_pUtils;
static char g_exceptionBuffer[8192];

///////////////////////////////
// User editing starts here
///////////////////////////////

RCINT64 GetMinorVersionWeCanRead()
{
    return 10; /* e.g. version 3.10 */
}

#define MAX_RECENTITYCODES 1000
#define MAX_EXCHANGERATES 100

void OutputTree(FILE *fp, RCHANDLE treeRoot)
{
    // Actually, for this example we'll just output the MOC calls from the tree

    HEADER hdr;
    RC_VISIT_HANDLE mocVisit, tapDecimalPlacesVisit;
    RCHANDLE moc, tdp;
    RCINT64 decimals;
    char *mscIds[MAX_RECENTITYCODES];
    int totalMSCs = 0;
    double exchangeRates[MAX_EXCHANGERATES];
    int totalERs = 0;
    int bytesRead = -1;

    memset(mscIds, 0, sizeof(mscIds));
    memset(exchangeRates, 0, sizeof(exchangeRates));

    // First, the HEADER block
    memset(&hdr, 0, sizeof(hdr));

    // SENDER
    {
	char tmp[sizeof(hdr.SENDER)+1];
	memset(&hdr.SENDER, ' ', sizeof(hdr.SENDER)-1);
	CONTROLLED_RUN(
	    g_pUtils->pfnGetString( treeRoot, "transferBatch.batchControlInfo.sender", tmp, sizeof(tmp), &bytesRead) )
	memcpy(&hdr.SENDER, tmp, bytesRead>(sizeof(hdr.SENDER)-1)?(sizeof(hdr.SENDER)-1):bytesRead);
	hdr.bSENDER_Exists = 1;
    }
    // RECIPIENT
    {
	char tmp[sizeof(hdr.RECIPIENT)+1];
	memset(&hdr.RECIPIENT, ' ', sizeof(hdr.RECIPIENT)-1);
	CONTROLLED_RUN(
	    g_pUtils->pfnGetString( treeRoot, "transferBatch.batchControlInfo.recipient", tmp, sizeof(tmp), &bytesRead) )
	memcpy(&hdr.RECIPIENT, tmp, bytesRead>(sizeof(hdr.RECIPIENT)-1)?(sizeof(hdr.RECIPIENT)-1):bytesRead);
	hdr.bRECIPIENT_Exists = 1;
    }
    // FSN
    {
	char tmp[sizeof(hdr.FSN)+1];
	memset(&hdr.FSN, ' ', sizeof(hdr.FSN)-1);
	CONTROLLED_RUN(
	    g_pUtils->pfnGetString( treeRoot, "transferBatch.batchControlInfo.fileSequenceNumber", tmp, sizeof(tmp), &bytesRead) )
	memcpy(&hdr.FSN, tmp, bytesRead>(sizeof(hdr.FSN)-1)?(sizeof(hdr.FSN)-1):bytesRead);
	hdr.bFSN_Exists = 1;
    }
    // FileCreationDate
    {
	char tmp[20];
	if (0 == g_pUtils->pfnGetString(treeRoot, "transferBatch.batchControlInfo.fileCreationTimeStamp.localTimeStamp", tmp, sizeof(tmp), &bytesRead)) {
	    if (bytesRead<=(sizeof(tmp)-1))
		tmp[bytesRead] = 0;
	    else
		tmp[sizeof(tmp)-1] = 0;
	    if (strptime(tmp, "%Y%m%d%H%M%S", &hdr.FILE_CREATION_DATE))
		hdr.bFILE_CREATION_DATE_Exists = 1;
	}
    }
    OutputHEADER(fp, &hdr);

    // In preparation for MSCID and EXCHANGE_RATE, we'll build the lookup tables
    // First, MSCIDs
    {
	RCHANDLE networkInfo;
	if (0 == g_pUtils->pfnNavigateAndObtainHandle( treeRoot, "transferBatch.networkInfo", &networkInfo)) {
	    RC_VISIT_HANDLE visit = g_pUtils->pfnInitializeVisit(networkInfo, "RecEntityInformation");
	    RCHANDLE rei;
	    while (rei = g_pUtils->pfnGetNext(visit)) {
		RCINT64 code;
		if (0 == g_pUtils->pfnGetInteger(rei, "recEntityCode", &code)) {
		    char tmp[16];
		    if (0 == g_pUtils->pfnGetString(rei, "recEntityId.mscId", tmp, sizeof(tmp), &bytesRead)) {
			if (code >= MAX_RECENTITYCODES) {
			    SNPRINTF(g_exceptionBuffer, sizeof(g_exceptionBuffer), "Up to %d recEntityCodes supported.\n", MAX_RECENTITYCODES);
			    g_pUtils->pfnThrowException(g_exceptionBuffer);
			} else
			    tmp[bytesRead] = 0;
			    mscIds[code] = strdup(tmp);
		    }
		}
	    }
	}
    }
    // Second: EXCHANGE_RATEs
    {
	RCHANDLE accountingInfo;
	if (0 == g_pUtils->pfnNavigateAndObtainHandle( treeRoot, "transferBatch.accountingInfo", &accountingInfo)) {
	    RC_VISIT_HANDLE visit = g_pUtils->pfnInitializeVisit(accountingInfo, "CurrencyConversion");
	    RCHANDLE cc;
	    while (cc = g_pUtils->pfnGetNext(visit)) {
		RCINT64 code;
		if (0 == g_pUtils->pfnGetInteger(cc, "exchangeRateCode", &code)) {
		    RCINT64 edecimals, rate;
		    if (0 == g_pUtils->pfnGetInteger(cc, "numberOfDecimalPlaces", &edecimals) &&
		       (0 == g_pUtils->pfnGetInteger(cc, "exchangeRate", &rate))) 
		    {
			if (code >= MAX_EXCHANGERATES) {
			    SNPRINTF(g_exceptionBuffer, sizeof(g_exceptionBuffer), "Up to %d exchangeRateCodes supported.\n", MAX_EXCHANGERATES);
			    g_pUtils->pfnThrowException(g_exceptionBuffer);
			} else
			    exchangeRates[code] = ((double)rate)/pow(10.0, (double) edecimals);
		    }
		}
	    }
	}
    }

    // Learn the TAP decimal places
    tapDecimalPlacesVisit = g_pUtils->pfnInitializeVisit( treeRoot, "TapDecimalPlaces" );
    while((tdp = g_pUtils->pfnGetNext(tapDecimalPlacesVisit))) {
        CONTROLLED_RUN( g_pUtils->pfnGetInteger(tdp, "", &decimals) )
        g_pUtils->pfnDestroySearch(tapDecimalPlacesVisit);
        break;
    }

    // And then, for each MOC call, one MOC block
    mocVisit = g_pUtils->pfnInitializeVisit( treeRoot, "MobileOriginatedCall" );
    while((moc = g_pUtils->pfnGetNext(mocVisit))) {
        MOC mocData;
        memset(&mocData, 0, sizeof(mocData));

        // IMSI
        {
            char tmp[sizeof(mocData.IMSI)+1];
            memset(&mocData.IMSI, ' ', sizeof(mocData.IMSI)-1);
            CONTROLLED_RUN(
                g_pUtils->pfnGetString(
                    moc,
                    "basicCallInformation.chargeableSubscriber.simChargeableSubscriber.imsi",
                    tmp,
                    sizeof(tmp),
		    &bytesRead) )
            memcpy(&mocData.IMSI, tmp, bytesRead>(sizeof(mocData.IMSI)-1)?(sizeof(mocData.IMSI)-1):bytesRead);
            mocData.bIMSI_Exists = 1;
        }

        // CALLED_NUMBER
        {
            char tmp[sizeof(mocData.CALLED_NUMBER)+1];
            memset(&mocData.CALLED_NUMBER, ' ', sizeof(mocData.CALLED_NUMBER)-1);
            CONTROLLED_RUN(
                g_pUtils->pfnGetString(
                    moc,
                    "basicCallInformation.destination.calledNumber",
                    tmp,
                    sizeof(tmp),
		    &bytesRead) )
            memcpy(&mocData.CALLED_NUMBER, tmp, bytesRead>(sizeof(mocData.CALLED_NUMBER)-1)?(sizeof(mocData.CALLED_NUMBER)-1):bytesRead);
            mocData.bCALLED_NUMBER_Exists = 1;
        }

        // NATIONAL_IND
        {
            RCINT64 modInd;
            if (0 == g_pUtils->pfnGetInteger(moc, "basicServiceUsedList.[0].chargeInformationList.[0].callTypeGroup.callTypeLevel1", &modInd)) {
                mocData.NATIONAL_IND = (int) modInd;
                mocData.bNATIONAL_IND_Exists = 1;
            }
        }

        // CALL_DATE_TIME
        {
            char tmp[20];
            if (0 == g_pUtils->pfnGetString(moc, "basicCallInformation.callEventStartTimeStamp.localTimeStamp", tmp, sizeof(tmp), &bytesRead)) {
		if (bytesRead<=(sizeof(tmp)-1))
		    tmp[bytesRead] = 0;
		else
		    tmp[sizeof(tmp)-1] = 0;
                if (strptime(tmp, "%Y%m%d%H%M%S", &mocData.CALL_DATE_TIME))
                    mocData.bCALL_DATE_TIME_Exists = 1;
            }
        }

        // DURATION
        {
            RCINT64 duration;
            if (0 == g_pUtils->pfnGetInteger(moc, "basicCallInformation.totalCallEventDuration", &duration)) {
                mocData.DURATION = (int) duration;
                mocData.bDURATION_Exists = 1;
            }
        }

	// MSCID
	{
	    // To get to the (possible) msc, we'll get the recEntityCode
	    RCINT64 code;
	    if (0 == g_pUtils->pfnGetInteger(moc, "locationInformation.networkLocation.recEntityCode", &code)) {
		if (code<MAX_RECENTITYCODES && mscIds[code] != NULL) {
		    memcpy(&mocData.MSCID, mscIds[code], (strlen(mscIds[code])>(sizeof(mocData.MSCID)-1))?(sizeof(mocData.MSCID)-1):strlen(mscIds[code]));
		    mocData.bMSCID_Exists = 1;
		}
	    }
	}

	// CALLEDCOUNTRYCODE
        {
            char tmp[sizeof(mocData.CALLEDCOUNTRYCODE)+1];
            memset(&mocData.CALLEDCOUNTRYCODE, ' ', sizeof(mocData.CALLEDCOUNTRYCODE)-1);
            if (0 == g_pUtils->pfnGetString(
		moc,
		"basicServiceUsedList.[0].chargeInformationList.[0].callTypeGroup.calledCountryCode",
		tmp,
		sizeof(tmp),
		&bytesRead) )
	    {
		memcpy(&mocData.CALLEDCOUNTRYCODE, tmp, bytesRead>(sizeof(mocData.CALLEDCOUNTRYCODE)-1)?(sizeof(mocData.CALLEDCOUNTRYCODE)-1):bytesRead);
		mocData.bCALLEDCOUNTRYCODE_Exists = 1;
	    }
        }
	
        // CHARGE and CHARGEABLE_UNITS
        {
            RC_VISIT_HANDLE visit = g_pUtils->pfnInitializeVisit(moc, "ChargeDetail");
            RCHANDLE cd;
            while (cd = g_pUtils->pfnGetNext(visit)) {
                char chargeType[6];
                if (0 == g_pUtils->pfnGetString(cd, "chargeType", chargeType, sizeof(chargeType), &bytesRead)) {
		    if (bytesRead<=(sizeof(chargeType)-1))
			chargeType[bytesRead] = 0;
		    else
			chargeType[sizeof(chargeType)-1] = 0;
                    if (!strcmp(chargeType, "00")) {
                        RCINT64 charge, chargeableUnits;
                        if (0 == g_pUtils->pfnGetInteger(cd, "charge", &charge)) {
                            mocData.CHARGE = ((double) charge)/pow(10.0, (double)decimals);
                            mocData.bCHARGE_Exists = 1;
                        }
                        if (0 == g_pUtils->pfnGetInteger(cd, "chargeableUnits", &chargeableUnits)) {
                            mocData.CHARGEABLE_UNITS = chargeableUnits;
                            mocData.bCHARGEABLE_UNITS_Exists = 1;
                        }
			g_pUtils->pfnDestroySearch(visit);
			break;
                    }
                }
            }
        }

	// EXCHANGE_RATE
	{
	    RCINT64 code;
	    if (0 == g_pUtils->pfnGetInteger(moc, "basicServiceUsedList.[0].chargeInformationList.[0].exchangeRateCode", &code)) {
		if (code<MAX_EXCHANGERATES && exchangeRates[code] != 0.0) {
		    mocData.EXCHANGE_RATE = exchangeRates[code];
		} else {
		    mocData.EXCHANGE_RATE = 1.0;
		}
	    } else {
		mocData.EXCHANGE_RATE = 1.0;
	    }
	    mocData.bEXCHANGE_RATE_Exists = 1;
	}

        OutputMOC(fp, &mocData);
    }
}

///////////////////////////////
// User editing stops here
///////////////////////////////
